/*	Copyright (C) 1992 Free Software Foundation, Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this software; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


#include "map.h"
#include "font.h"
#include "ctype.h"
#include "math.h"
#include "window.h"
#include "io_abstract.h"

static struct font_memo default_cell_font =
{
  "*times-medium-r-*",
  "Times-Roman",
  1.0,
  0
};

struct font_memo *font_list = 0;
static xx_Map font_map = 0;

static void 
font_memo_copy (leaf, rect, data)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     void *data;
{
  leaf->data = data;
}

static void 
font_memo_noop (data)
     void *data;
{
}

void 
init_fonts ()
{
  font_map = xx_map (font_memo_copy, font_memo_noop);
}


char *default_x_name = "*times-medium-r-*";
char *default_ps_name = "Times-Roman";
extern int errno;

struct font_names
{
  struct font_names * next;
  char * name;
  char * definition;
};

static struct font_names * font_names;

void
define_font (description)
     char * description;
{
  char * p;
  char * p2;
  char oleo_name[1000];
  struct font_names * fn;

  for (p = description; isspace (*p); ++p);
  for (p2 = oleo_name; *p && (*p != ',') && !isspace (*p); *p2++ = *p++);
  *p2 = '\0';
  if (oleo_name[0] == '\0')
    return;
  
  for (fn = font_names; fn; fn = fn->next)
    if (!strcmp(fn->name, oleo_name))
      {
	free (fn->definition);
	fn->definition = 0;
      }
  if (!fn)
    {
      fn = (struct font_names *)ck_malloc (sizeof (*fn));
      fn->name = strdup (oleo_name);
      fn->next = font_names;
      font_names = fn;
    }
  fn->definition = strdup (p);
}

struct font_memo *
intern_font (fullname)
     char *fullname;
{
  struct font_memo *it = font_list;
  char name[1000];
  char psname[1000];
  char retrybuf[1000];
  char *p;
  char *p2;
  double scale;

 retry:
  for (p = fullname; isspace (*p); ++p);
  for (p2 = name; *p && (*p != ',') && !isspace (*p); *p2++ = *p++);
  *p2 = '\0';
  if (name[0] == '\0')
    strcpy (name, default_x_name);

  /* The X name might really have been a `defined font name'.
   * In that case, we splice the definition in in place of the name
   * and start over again.
   */
  {
    struct font_names * fn;
    for (fn = font_names; fn; fn = fn->next)
      if (!strcmp (name, fn->name))
	{
	  strcpy (retrybuf, fn->definition);
	  strcat (retrybuf, p);
	  fullname = retrybuf;
	  goto retry;
	}
  }

  while (isspace (*p))
    ++p;
  if (*p == ',')
    ++p;
  while (isspace (*p))
    ++p;
  for (p2 = psname; *p && (*p != ',') && !isspace (*p); *p2++ = *p++);
  *p2 = '\0';
  if (psname[0] == '\0')
    strcpy (psname, default_ps_name);
  while (isspace (*p))
    ++p;
  if (*p == ',')
    ++p;
  while (isspace (*p))
    ++p;
  if (isdigit (*p))
    {
      errno = 0;
      scale = atof (p);
      if (errno)
	scale = 1.;
    }
  else
    scale = 1.;
  while (it)
    {
      if (scale > it->scale)
	{
	  it = 0;
	  break;
	}
      if (scale == it->scale)
	{
	  int cmp = strcmp (name, it->name);
	  if (cmp > 0)
	    {
	      it = 0;
	      break;
	    }
	  if (!cmp)
	    {
	      cmp = strcmp (psname, it->psname);
	      if (cmp > 0)
		{
		  it = 0;
		  break;
		}
	      if (!cmp)
		break;
	    }
	}
      it = it->next;
    }
  if (!it)
    {
      struct font_memo *f =
	(struct font_memo *) ck_malloc (sizeof (*f));
      f->scale = scale;
      f->name = strdup (name);
      f->psname = strdup (psname);
      f->next = font_list;
      font_list = f;
      it = f;
    }
  return it;
}

static int 
write_font (out, a, b)
     void **out;
     void *a;
     void *b;
{
  if ((a == b) || !b)
    {
      *out = a;
      return !!a;
    }
  else
    {
      *out = b;
      return 1;
    }
}

void 
set_area_font (rng, font)
     struct rng *rng;
     struct font_memo *font;
{
  struct xx_sIntRectangle rect;

  rect.x = rng->lr;
  rect.y = rng->lc;
  rect.w = rng->hr - rng->lr + 1;
  rect.h = rng->hc - rng->lc + 1;
  xx_map_write_rectangle (font_map, &rect, write_font, (void *) font);
  io_redo_region (rng);
}

void 
flush_fonts ()
{
  xx_map_clear (font_map);
}

struct qframe
{
  font_iterator iter;
  void *vframe;
};

static int 
doquery (leaf, rect, vqframe)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     void *vqframe;
{
  struct qframe *qf = (struct qframe *) vqframe;
  struct rng rng;
  rng.lr = rect->x;
  rng.lc = rect->y;
  rng.hr = rect->w + rect->x - 1;
  rng.hc = rect->h + rect->y - 1;

  return qf->iter (&rng, (struct font_memo *) leaf->data, qf->vframe);
}


void 
query_fonts (iter, rng, vframe)
     font_iterator iter;
     struct rng *rng;
     void *vframe;
{
  struct xx_sIntRectangle rect;
  struct qframe qf;
  qf.iter = iter;
  qf.vframe = vframe;

  rect.x = rng->lr;
  rect.y = rng->lc;
  rect.w = rng->hr - rng->lr + 1;
  rect.h = rng->hc - rng->lc + 1;
  xx_map_query (font_map, &rect, doquery, (void *) &qf);
}

static int 
get_one_font (leaf, rect, vanswer)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     void *vanswer;
{
  *((void **) vanswer) = leaf->data;
  return 0;
}

struct font_memo *
get_font (r, c)
     CELLREF r;
     CELLREF c;
{
  struct xx_sIntRectangle rect;
  struct font_memo *answer = 0;
  rect.x = r;
  rect.w = 1;
  rect.y = c;
  rect.h = 1;
  xx_map_query (font_map, &rect, get_one_font, (void *) &answer);
  if (answer == 0)
    answer = &default_cell_font;
  return answer;
}

struct font_memo * 
default_font ()
{
  return &default_cell_font;
}


void
set_x_default_font (str)
     char * str;
{
  default_cell_font.name = strdup(str);
  io_repaint ();
}

void
set_ps_font_cmd (name)
     char * name;
{
  /* A deliberate core leak so we can init with a constant. */
  default_cell_font.psname = strdup (name);
}
