/*	Copyright (C) 1992 Free Software Foundation, Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this software; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


#include "map.h"
#include "irpile.h"

xx_Map
xx_map (copy, destroy)
     xx_MapCopier copy;
     xx_MapDestructor destroy;
{
  xx_Map result = (xx_Map) ck_malloc (sizeof (*result));
  result->copy = copy;
  result->destroy = destroy;
  result->rtree = xx_rtree ();
  return result;
}

void
xx_free_map (map)
     xx_Map map;
{
  xx_map_clear (map);
  xx_free_rtree (map->rtree);
  ck_free (map);
}

int
xx_map_empty (map)
     xx_Map map;
{
  return xx_rtree_empty (map->rtree);
}

static int
deletor (leaf, rect, voidmap)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     void *voidmap;
{
  xx_Map map = (xx_Map) voidmap;

  map->destroy (leaf->data);
  return 1;
}

void
xx_map_clear (map)
     xx_Map map;
{
  xx_rtree_query (map->rtree, &map->rtree->domain, deletor, (void *) map);
  xx_rtree_clear (map->rtree);
}


static int
delete_combinor (out, a, b)
     void **out;
     void *a;
     void *b;
{
  *out = 0;
  return 0;
}

void
xx_map_clear_rectangle (map, ir)
     xx_Map map;
     xx_IntRectangle ir;
{
  xx_map_write_rectangle (map, ir, delete_combinor, 0);
}

void
xx_map_query (map, rect, fn, data)
     xx_Map map;
     xx_IntRectangle rect;
     xx_LoopFn fn;
     void *data;
{
  xx_rtree_query (map->rtree, rect, fn, data);
}

struct write_frame
{
  xx_IntRectangle rect;
  xx_IntRectanglePile hitrects;
  xx_RtreeLeafPile hitleaves;
};

static int
write_rectangle (leaf, rect, voidframe)
     xx_RtreeLeaf leaf;
     xx_IntRectangle rect;
     void *voidframe;
{
  struct write_frame *frame = (struct write_frame *) voidframe;
  xx_IntRectangle outrect = xx_intrectanglepile_reserve (frame->hitrects, 1);
  xx_rtreeleafpile_add (frame->hitleaves, leaf);
  *outrect = *rect;
  return 1;
}

void
xx_map_write_rectangle (map, rect, combinor, data)
     xx_Map map;
     xx_IntRectangle rect;
     xx_MapCombinor combinor;
     void *data;
{
  struct write_frame frame;
  static xx_IntRectanglePile irpile = 0;
  static xx_RtreeLeafPile rlpile = 0;

  if (!irpile)
    {
      irpile = xx_intrectanglepile ();
      rlpile = xx_rtreeleafpile ();
    }
  frame.rect = rect;
  frame.hitrects = irpile;
  frame.hitleaves = rlpile;

  xx_rtree_query (map->rtree, rect, write_rectangle, &frame);

  {
    int x;
    xx_IntRectangle hitrects = xx_intrectanglepile_read (frame.hitrects);
    xx_RtreeLeaf *hitleaves = xx_rtreeleafpile_read (frame.hitleaves);
    void *newval;
    int include_new;

    include_new = combinor (&newval, 0, data);


    for (x = 0; x < frame.hitleaves->used; ++x)
      {
	int are_leftovers;
	int include_old, include_intersection;
	void *intersection_becomes;
	void *old_becomes;
	void *old = hitleaves[x]->data;

	struct xx_sIntRectangle intersection;

	intersection = hitrects[x];
	xx_IRclip (&intersection, rect);

	include_intersection =
	  combinor (&intersection_becomes, hitleaves[x]->data, data);

	include_old =
	  combinor (&old_becomes, hitleaves[x]->data, 0);

	if (include_old && include_intersection
	    && intersection_becomes == old_becomes)
	  {
	    map->copy (hitleaves[x], &hitrects[x], old_becomes);
	  }
	else if (include_intersection)
	  {
	    if (xx_IRarea (&intersection) != xx_IRarea (&hitrects[x]))
	      {
		map->copy (hitleaves[x], &intersection, intersection_becomes);
		xx_rtree_delete (map->rtree, hitleaves[x]);
		xx_rtree_add (map->rtree, &intersection, hitleaves[x]);
		are_leftovers = 1;
	      }
	    else
	      {
		are_leftovers = 0;
		if (intersection_becomes == newval)
		  {
		    hitrects[x].w = hitrects[x].h = 0;
		    xx_rtree_delete (map->rtree, hitleaves[x]);
		    xx_free_rtreeleaf (hitleaves[x]);
		  }
		else
		  map->copy (hitleaves[x], &intersection,
			     intersection_becomes);
	      }
	  }
	else
	  {
	    xx_rtree_delete (map->rtree, hitleaves[x]);
	    xx_free_rtreeleaf (hitleaves[x]);
	    are_leftovers =
	      xx_IRarea (&intersection) != xx_IRarea (&hitrects[x]);
	  }

	if (are_leftovers && include_old)
	  {
	    int y;
	    struct xx_sIntRectangle oldrects[4];
	    int leftovers = xx_IRsubtract (oldrects,
					   &hitrects[x], &intersection);

	    for (y = 0; y < leftovers; ++y)
	      {
		xx_RtreeLeaf newleaf = xx_rtreeleaf (0);
		map->copy (newleaf, &oldrects[y], old_becomes);
		xx_rtree_add (map->rtree, &oldrects[y], newleaf);
	      }
	  }
	map->destroy (old);
	if (include_old)
	  map->destroy (old_becomes);
	if (include_intersection)
	  map->destroy (intersection_becomes);
      }


    if (include_new)
      {
	int y;
	static xx_IntRectanglePile srcrects = 0;
	static xx_IntRectanglePile temp = 0;
	if (!srcrects)
	  {
	    srcrects = xx_intrectanglepile ();
	    temp = xx_intrectanglepile ();
	  }

	*xx_intrectanglepile_reserve (srcrects, 1) = *rect;

	for (y = 0; y < frame.hitrects->used; ++y)
	  {
	    xx_IntRectangle diff = xx_intrectanglepile_reserve (temp, 4);
	    xx_IntRectangle src = xx_intrectanglepile_read (srcrects);
	    int z;

	    for (z = 0; z < srcrects->used; ++z)
	      if (hitrects[y].w)
		{
		  int frags = xx_IRsubtract (diff, src + z, hitrects + y);
		  diff = (xx_intrectanglepile_reserve (temp, frags)
			  - 4 + frags);
		  {
		    xx_IntRectanglePile t = srcrects;
		    srcrects = temp;
		    temp = t;
		    xx_intrectanglepile_clear (temp);
		    xx_intrectanglepile_unreserve (srcrects, 4);
		  }
		}
	  }


	{
	  xx_IntRectangle todo = xx_intrectanglepile_read (srcrects);
	  xx_intrectanglepile_clear (temp);
	  for (y = 0; y < srcrects->used; ++y)
	    {
	      xx_RtreeLeaf newleaf = xx_rtreeleaf (0);
	      map->copy (newleaf, todo + y, newval);
	      xx_rtree_add (map->rtree, todo + y, newleaf);
	    }
	  xx_intrectanglepile_clear (srcrects);
	}
	map->destroy (newval);
      }
  }
  xx_rtreeleafpile_clear (frame.hitleaves);
  xx_intrectanglepile_clear (frame.hitrects);
}
